package main

import (
	"fmt"
	"strings"
	"time"
)

type ChSpan struct {
}

type TopoData struct {
	TxServiceName   string
	TxEndpoint      string
	TxCmdbID        string
	SrcServiceName  string
	SrcEndpoint     string
	SrcCmdbID       string
	DestServiceName string
	DestEndpoint    string
	DestCmdbID      string
	RequestCount    int64
	FailedCount     int64
	AvgResponseTime int64
}

func (t *TopoData) TopoID() string {
	return fmt.Sprintf("%s.%s-%s.%s-%s.%s", t.TxServiceName, t.TxEndpoint,
		t.SrcServiceName, t.SrcEndpoint, t.DestServiceName, t.DestEndpoint)
}

type MetricsMeters struct {
	requestCount          int64
	failedCount           int64
	exceptionCount        int64
	aggDuration           time.Duration
	responseTimeSumNS     int64
	apdexGoodRequestCount int64
	apdexFairRequestCount int64
}

func (m MetricsMeters) GetCallPerMinute() int64 {
	minutes := m.aggDuration.Minutes()
	if minutes == 0 {
		return 0
	}
	return int64(float64(m.requestCount) / minutes)
}

func (m MetricsMeters) GetCallCount() int64 {
	return m.requestCount
}

func (m MetricsMeters) GetFailedCount() int64 {
	return m.failedCount
}

func (m MetricsMeters) GetExceptionCount() int64 {
	return m.exceptionCount
}

func (m MetricsMeters) GetAvgResponseTimeMS() int64 {
	if m.requestCount == 0 {
		return 0
	}
	return (m.responseTimeSumNS / m.requestCount) / 1e6
}

func (m MetricsMeters) GetApdex() float64 {
	if m.requestCount == 0 {
		return 0.0
	}
	return float64(m.apdexGoodRequestCount+m.apdexFairRequestCount/2) / float64(m.requestCount)
}

func (m MetricsMeters) GetSlowCount() int64 {
	return 0
}

func (m MetricsMeters) GetStatementCount() int64 {
	return 0
}

type ServiceInstanceEndpointMetrics struct {
	Endpoint string
	Metrics  MetricsMeters
}

type ServiceInstanceMetric struct {
	InstanceID      string
	InstanceIP      string
	EndpointMetrics []ServiceInstanceEndpointMetrics
}

func (m ServiceInstanceMetric) GetCPM() int64 {
	sum := int64(0)
	count := int64(len(m.EndpointMetrics))
	if count == 0 {
		return 0
	}
	for _, ep := range m.EndpointMetrics {
		sum += ep.Metrics.GetCallPerMinute()
	}
	return sum / count
}

func (m ServiceInstanceMetric) GetCallCount() int64 {
	cc := int64(0)
	for _, metric := range m.EndpointMetrics {
		cc += metric.Metrics.GetCallCount()
	}
	return cc
}

func (m ServiceInstanceMetric) GetFailCount() int64 {
	cc := int64(0)
	for _, metric := range m.EndpointMetrics {
		cc += metric.Metrics.GetFailedCount()
	}
	return cc
}

func (m ServiceInstanceMetric) GetExceptionCount() int64 {
	cc := int64(0)
	for _, metric := range m.EndpointMetrics {
		cc += metric.Metrics.GetExceptionCount()
	}
	return cc
}

func (m ServiceInstanceMetric) GetAvgResponseTimeMS() int64 {
	if len(m.EndpointMetrics) == 0 {
		return 0
	}
	sum := int64(0)
	count := int64(len(m.EndpointMetrics))
	for _, ep := range m.EndpointMetrics {
		sum += ep.Metrics.GetAvgResponseTimeMS()
	}
	return sum / count
}

func (m ServiceInstanceMetric) GetApdex() float64 {
	sum := float64(0)
	if len(m.EndpointMetrics) == 0 {
		return 1.0
	}
	for _, metric := range m.EndpointMetrics {
		sum += metric.Metrics.GetApdex()
	}
	return sum / float64(len(m.EndpointMetrics))
}

type ServiceMetricData struct {
	AppAlias        string
	CmdbID          string
	ServiceName     string
	InstanceMetrics []ServiceInstanceMetric
}

func (m ServiceMetricData) GetCPM() int64 {
	if len(m.InstanceMetrics) == 0 {
		return 0
	}
	sum := int64(0)
	count := int64(len(m.InstanceMetrics))
	for _, ep := range m.InstanceMetrics {
		sum += ep.GetCPM()
	}
	return sum / count
}

func (m ServiceMetricData) GetCallCount() int64 {
	cc := int64(0)
	for _, metric := range m.InstanceMetrics {
		cc += metric.GetCallCount()
	}
	return cc
}

func (m ServiceMetricData) GetFailCount() int64 {
	cc := int64(0)
	for _, metric := range m.InstanceMetrics {
		cc += metric.GetFailCount()
	}
	return cc
}

func (m ServiceMetricData) GetExceptionCount() int64 {
	cc := int64(0)
	for _, metric := range m.InstanceMetrics {
		cc += metric.GetExceptionCount()
	}
	return cc
}

func (m ServiceMetricData) GetAvgResponseTimeMS() int64 {
	if len(m.InstanceMetrics) == 0 {
		return 0
	}
	sum := int64(0)
	count := int64(len(m.InstanceMetrics))
	for _, ep := range m.InstanceMetrics {
		sum += ep.GetAvgResponseTimeMS()
	}
	return sum / count
}

func (m ServiceMetricData) GetApdex() float64 {
	if len(m.InstanceMetrics) == 0 {
		return 1
	}
	sum := float64(0)
	count := int64(len(m.InstanceMetrics))
	for _, ep := range m.InstanceMetrics {
		sum += ep.GetApdex()
	}
	return sum / float64(count)
}

func (m ServiceMetricData) GetEndPointCount() int {
	s := 0
	for _, instanceMetric := range m.InstanceMetrics {
		s += len(instanceMetric.EndpointMetrics)
	}
	return s
}

func (m *ServiceMetricData) AddMetric(instance string, em ServiceInstanceEndpointMetrics) {
	for i, instanceMetric := range m.InstanceMetrics {
		if instanceMetric.InstanceID == instance {
			m.InstanceMetrics[i].EndpointMetrics = append(m.InstanceMetrics[i].EndpointMetrics, em)
			return
		}
	}
	m.InstanceMetrics = append(m.InstanceMetrics, ServiceInstanceMetric{
		InstanceID:      instance,
		InstanceIP:      instance,
		EndpointMetrics: []ServiceInstanceEndpointMetrics{em},
	})
}

func (m *ServiceMetricData) GetEndPointsStr() string {
	ma := make(map[string]bool)
	es := make([]string, 0)
	for _, instanceMetric := range m.InstanceMetrics {
		for _, metric := range instanceMetric.EndpointMetrics {
			_, find := ma[metric.Endpoint]
			if !find {
				es = append(es, metric.Endpoint)
				ma[metric.Endpoint] = true
			}
		}
	}
	if len(es) == 0 {
		return ""
	}
	return strings.Join(es, ",")
}
